import importlib.metadata
import logging
import os
import shutil
from typing import Dict, Any, List

import click
from sqlalchemy import text

from dbd.log.dbd_exception import DbdException
from dbd.config.dbd_profile import DbdProfile
from dbd.config.dbd_project import DbdProject
from dbd.executors.model_executor import ModelExecutor
from dbd.log.dbd_logger import setup_logging

log = logging.getLogger(__name__)

this_script_dir = os.path.dirname(__file__)


class Dbd(object):
    """
    Top level CLI object
    """

    def __init__(self, debug: bool = False, logfile: str = 'dbd.log', profile: str = 'dbd.profile',
                 project: str = 'dbd.project'):
        """
        Constructor
        :param bool debug: debug flag
        :param str logfile: log file
        :param str profile: profile file
        :param str project: project file
        """
        self.__debug = debug
        self.__logfile = logfile
        self.__profile = profile
        self.__project = project

    def debug(self) -> bool:
        """
        Debug flag getter
        :return: debug flag
        :rtype: bool
        """
        return self.__debug

    def logfile(self) -> str:
        """
        Logfile getter
        :return: logfile
        :rtype: str
        """
        return self.__logfile

    def profile(self) -> str:
        """
        Profile getter
        :return: profile
        :rtype: str
        """
        return self.__profile

    def project(self) -> str:
        """
        Project getter
        :return: project
        :rtype: str
        """
        return self.__project


def print_version():
    """
    Prints DBD version
    """
    click.echo(f"You're using DBD version {importlib.metadata.version('dbd')}.")


@click.group(invoke_without_command=True)
@click.option('--debug/--no-debug', envvar='DBD_DEBUG', default=False, help='Sets debugging on/off')
@click.option('--version', help="Print the DBD version and exit.", is_flag=True, is_eager=True)
@click.option('--logfile', envvar='DBD_LOG_FILE', default='dbd.log', help='Log file location')
@click.option('--profile', envvar='DBD_PROFILE', default='dbd.profile', help='Profile configuration file')
@click.option('--project', envvar='DBD_PROJECT', default='dbd.project', help='Project configuration file')
@click.pass_context
def cli(ctx, debug, logfile, version, profile, project):
    if debug:
        click.echo(f"Logging DEBUG info to '{logfile}'")
        setup_logging(logging.DEBUG, logfile)
    if version:
        print_version()
        ctx.exit(0)
    ctx.obj = Dbd(debug, logfile, profile, project)


@cli.command(help='Initializes a new DBD project.')
@click.argument('dest', required=False, default='my_new_dbd_project')
@click.pass_obj
def init(dbd, dest):
    try:
        src = os.path.join(this_script_dir, '..', 'resources', 'template')
        if os.path.exists(dest):
            log.error(f"Can't overwrite directory '{dest}'")
            raise DbdException(f"Can't overwrite directory '{dest}'")
        shutil.copytree(src, dest)
        click.echo(f"New project {dest} generated. Do cd {dest}; dbd run .")
    except DbdException as d:
        click.echo(f"ERROR: '{d}'")


@cli.command(help='Executes project.')
@click.argument('dest', required=False, default='.')
@click.pass_obj
def run(dbd, dest):
    try:
        prf = DbdProfile.load(os.path.join('.', dbd.profile()))
        prj = DbdProject.load(prf, os.path.join(dest, dbd.project()))
        model = ModelExecutor(prj)
        engine = prj.alchemy_engine_from_project()
        model.execute(engine)
    except DbdException as d:
        click.echo(f"ERROR: '{d}'")


@cli.command(help='Validates project.')
@click.argument('dest', required=False, default='.')
@click.pass_obj
def validate(dbd, dest):
    try:
        prf = DbdProfile.load(os.path.join('.', dbd.profile()))
        prj = DbdProject.load(prf, os.path.join(dest, dbd.project()))
        model = ModelExecutor(prj)
        engine = prj.alchemy_engine_from_project()
        try:
            engine.execute(text("SELECT 1"))
        except Exception:
            click.echo(
                f"Can't connect to the target database. Check profile configuration in "
                f"'{os.path.normpath(os.path.join(dest, dbd.profile()))}'.")
        validation_result, validation_errors = model.validate()
        if validation_result:
            click.echo("No errors found. Model is valid.")
        else:
            click.echo("Model isn't valid. Please fix the following errors:")
            __echo_validation_errors(validation_errors)
    except DbdException as d:
        click.echo(f"ERROR: '{d}'")


def __echo_validation_errors(validation_errors: Dict[str, Any]):
    """
    Top level function for printing validation errors
    :param validation_errors:
    :return:
    """
    __echo_validation_level(validation_errors)


class InvalidValidationErrorStructure(DbdException):
    pass


def __echo_validation_level(level_validation_errors: Dict[str, Any], indent: int = 0):
    """
    Echo validation error line (called recursively on all Dict values)
    :param level_validation_errors: Dict with validation result
    :param indent: indentation level
    """
    for (k, v) in level_validation_errors.items():
        if isinstance(v, str):
            msg = f"{k}:{v}"
            click.echo(msg.rjust(indent * 2 + len(msg), ' '))
        elif isinstance(v, Dict):
            msg = f"{k}:"
            click.echo(msg.rjust(indent * 2 + len(msg), ' '))
            __echo_validation_level(v, indent + 1)
        elif isinstance(v, List):
            msg = f"{k}:{str(v)}"
            click.echo(msg.rjust(indent * 2 + len(msg), ' '))
        else:
            raise InvalidValidationErrorStructure(f"Invalid validation result: '{v}' isn't supported type.")



import logging

FORMAT = "%(asctime)s %(levelname)s %(module)s '%(message)s'"


def setup_logging(level: int, filename: str):
    """
    Sets up the logging
    :param int level: logging level
    :param str filename: log filename
    """
    logging.basicConfig(level=level, format=FORMAT, filename=filename)



import hashlib
from typing import Dict, List, Any, TypeVar

import sqlalchemy.engine
from sqlalchemy import select, Column
from sqlalchemy.testing.schema import Table

from dbd.db.db_table import DbTable
from dbd.tasks.db_table_task import DbTableTask
from dbd.utils.sql_parser import SqlParser
from dbd.utils.text_utils import fully_qualified_table_name

REFLECTION_TMP_VIEW_PREFIX = "tmp_view_"

EltTaskType = TypeVar('EltTaskType', bound='EltTask')


# noinspection PyMethodMayBeStatic
class EltTask(DbTableTask):
    """
    Database ELT task (INSERT from SELECT) definition
    """

    def __init__(self, task_def: Dict[str, Any]):
        """
        EltTask constructor
        :param Dict[str, Any] task_def: Target table definition
        """
        super().__init__(task_def)

    def __create_tmp_reflection_view(self, fully_qualified_view_name: str,
                                     alchemy_engine: sqlalchemy.engine.Engine):
        """
        Creates a temporary helper database view for introspecting the ELT task's
        SELECT result's structure (columns and types)
        TODO: some more clever way of introspection without creating a database view
        :param str fully_qualified_view_name: str fully qualified view name
        :param sqlalchemy.engine.Engine alchemy_engine: Engine SQLAlchemy Engine
        """
        view_sql_create = f"CREATE VIEW {fully_qualified_view_name} AS {self.sql()}"
        with alchemy_engine.connect() as conn:
            conn.execute(view_sql_create)

    def __drop_tmp_reflection_view(self, fully_qualified_view_name: str, alchemy_engine: sqlalchemy.engine.Engine):
        """
        Drops the temporary helper database view for introspecting the ELT task's
        SELECT result's structure (columns and types)
        TODO: some more clever way of introspection without creating a database view
        :param str fully_qualified_view_name: fully qualified view name
        :param sqlalchemy.engine.Engine alchemy_engine: Engine SQLAlchemy Engine
        """
        view_sql_drop = f"DROP VIEW IF EXISTS {fully_qualified_view_name}"
        with alchemy_engine.connect() as conn:
            conn.execute(view_sql_drop)

    def __sql_columns(self, target_alchemy_metadata: sqlalchemy.MetaData,
                      alchemy_engine: sqlalchemy.engine.Engine) -> List[sqlalchemy.Column]:
        """
        Introspects ETL task SELECT statement structure (result's column names and types)
        TODO: some more clever way of introspection without creating a database view
        :param sqlalchemy.MetaData target_alchemy_metadata: SQLAlchemy MetaData
        :param sqlalchemy.engine.Engine alchemy_engine: SQLAlchemy Engine
        :return: list of the SELECT result columns (SQLAlchemy Column[])
        :rtype: List[sqlalchemy.Column]
        """
        target_db_schema = self.target_schema()

        tmp_name = hashlib.md5(self.sql().encode('utf-8')).hexdigest()
        view_name = f"{REFLECTION_TMP_VIEW_PREFIX}{tmp_name}"
        fully_qualified_view_name = fully_qualified_table_name(target_db_schema, view_name)
        db_con = alchemy_engine.connect()
        self.__drop_tmp_reflection_view(fully_qualified_view_name, db_con)
        self.__create_tmp_reflection_view(fully_qualified_view_name, db_con)
        # SQLAlchemy reflection / autoload doesn't work with fully qualified view name
        tmp_reflection_view = Table(view_name, target_alchemy_metadata, autoload=True)
        self.__drop_tmp_reflection_view(fully_qualified_view_name, db_con)
        db_con.close()
        temp_view_select_all_columns = select(tmp_reflection_view.columns).select_from(tmp_reflection_view)
        return [Column(c.name, c.type) for c in temp_view_select_all_columns.selected_columns]

    # noinspection DuplicatedCode
    def __override_sql_column_definitions(self, target_alchemy_metadata: sqlalchemy.MetaData,
                                          alchemy_engine: sqlalchemy.engine.Engine) -> Dict[str, Any]:
        """
        Merges the SQL result column definitions with the column definitions from the ELT task.
        The column definitions override the introspected SQL types
        :param sqlalchemy.MetaData target_alchemy_metadata: MetaData SQLAlchemy MetaData
        :param sqlalchemy.engine.Engine alchemy_engine: Engine SQLAlchemy Engine
        :return: ETL task definition merged from the SELECT result structure overridden
            with the ELT task's explicit column definitions
        :rtype: Dict[str, Any]
        """
        table_def = self.table_def()
        column_overrides = table_def.get('columns', {})
        tmp_reflection_view_columns = self.__sql_columns(target_alchemy_metadata, alchemy_engine)
        ordered_columns = {}
        for c in tmp_reflection_view_columns:
            overridden_column = column_overrides.get(c.name)
            if overridden_column:
                if 'type' not in overridden_column:
                    overridden_column['type'] = c.type
                ordered_columns[c.name] = overridden_column
            else:
                ordered_columns[c.name] = {"type": c.type}

        table_def['columns'] = ordered_columns
        return table_def

    def sql(self) -> str:
        """
        Task SQL statement
        :return: task SQL statement
        """
        return self.task_data()

    def set_sql(self, sql: str):
        """
        Sets task SQL
        :param sql: SQL statement
        """
        self.set_task_data(sql)

    @classmethod
    def from_code(cls, task_def: Dict[str, Any]) -> EltTaskType:
        """
        Creates a new ELT task from table definition (dict)
        :param Dict[str, Any] task_def: ELT target table definition
        :return: new EltTask instance
        :rtype: EltTask
        """

        return EltTask(task_def)

    def create(self, target_alchemy_metadata: sqlalchemy.MetaData, alchemy_engine: sqlalchemy.engine.Engine):
        """
        Executes / creates the task
        :param sqlalchemy.MetaData target_alchemy_metadata: MetaData SQLAlchemy MetaData
        :param sqlalchemy.engine.Engine alchemy_engine:
        """
        process_def = self.process_def()
        materialization = process_def.get('materialization', 'table')
        if materialization == 'view':
            sql_text = f"CREATE VIEW  {self.fully_qualified_target()} AS {self.sql()}"
            with alchemy_engine.connect() as conn:
                conn.execute(sql_text)
        else:
            overridden_def = self.__override_sql_column_definitions(target_alchemy_metadata, alchemy_engine)
            db_table = DbTable.from_code(self.target(), overridden_def, target_alchemy_metadata, self.target_schema())
            self.set_db_table(db_table)
            db_table.create()
            columns = [c.name() for c in db_table.columns()]
            sql_text = f"INSERT INTO {self.fully_qualified_target()}({','.join(columns)}) {self.sql()}"
            with alchemy_engine.connect() as conn:
                conn.execute(sql_text)

    def depends_on(self) -> List[str]:
        """
        Returns list of tables that the ELT task depends on
        :return: list of tables that the ELT task depends on
        :rtype: List[str]
        """
        target_db_schema = self.target_schema()
        foreign_key_tables = super(EltTask, self).depends_on()
        sql_tables = SqlParser.extract_tables(self.sql())
        return foreign_key_tables + [
            f"{fully_qualified_table_name(target_db_schema, t)}" if len(t.split('.')) < 2 else t for t in sql_tables]



from typing import TypeVar

import sqlalchemy.engine
from sqlalchemy import text

from dbd.tasks.task import Task

DdlTaskType = TypeVar('DdlTaskType', bound='DdlTask')


class DdlTask(Task):
    """
    Task for executing SQL script
    """

    def __init__(self, target: str, target_schema: str):
        """
        Constructor
        :param str target: target (epilogue or prologue)
        :param str target_schema: Target schema for executing the SQL script
        """
        task_id = Task.generate_task_id(target, target_schema)
        super().__init__(task_id, target, target_schema)

    def sql_text(self) -> str:
        """
        Table definition dict
        :return: SQL script
        :rtype: str
        """
        return self.task_data()

    def set_sql_text(self, sql_script: str):
        """
        Sets SQL script (multiple SQL DDL statements separated by semicolon
        :param str sql_script: SQL script
        """
        self.set_task_data(sql_script)

    def create(self, target_alchemy_metadata: sqlalchemy.MetaData, alchemy_engine: sqlalchemy.engine.Engine):
        """
        Executes / creates the task
        :param sqlalchemy.MetaData target_alchemy_metadata: MetaData SQLAlchemy MetaData
        :param sqlalchemy.engine.Engine alchemy_engine:
        """
        for statement in self.sql_text():
            if statement:
                with alchemy_engine.connect() as conn:
                    conn.execute(text(statement))

    @classmethod
    def from_code(cls, target: str, target_schema: str) -> DdlTaskType:
        """
        Creates a new DDL task
        :param str target: target (epilogue or prologue)
        :param str target_schema: target schema
        :return: new DdlTask instance
        :rtype: DdlTask
        """

        return DdlTask(target, target_schema)



from typing import TypeVar, Dict, Any, Tuple

import sqlalchemy
from cerberus import Validator
from sqlalchemy import ForeignKey, Column

from dbd.generator.jinja_generator_env import JINJA_GENERATOR_ENV
from dbd.utils.sql_parser import SqlParser

DbColumnType = TypeVar('DbColumnType', bound='DbColumn')


class DbColumn:
    """
    Database column definition
    """

    def __init__(self, name: str, alchemy_column: sqlalchemy.Column):
        """
        Database column's constructor
        :param str name: column name
        :param sqlalchemy.Column alchemy_column: Column underlying SQLAlchemy column
        """
        self.__name = name
        self.__alchemy_column = alchemy_column

    def alchemy_column(self) -> sqlalchemy.Column:
        """
        :return: underlying SQLAlchemy column
        :rtype: sqlalchemy.Column
        """
        return self.__alchemy_column

    def name(self) -> str:
        """
        :return: column's name
        :rtype: str
        """
        return self.__name

    def code(self) -> str:
        """
        TODO: shall I turn this to JSON and to dict here?
        :return: DbColumn's code
        :rtype: str
        """
        template = JINJA_GENERATOR_ENV.get_template('column.j2')
        return template.render(dict(c=self.__alchemy_column))

    def __eq__(self, other: DbColumnType) -> bool:
        """
        Comparator
        :param DbColumn other: another DbColumn instance to compare this instance to
        :return: True if both DbColumns are equal, False otherwise
        :rtype: bool
        """
        if isinstance(other, DbColumn):
            return self.code() == other.code()

    @classmethod
    def from_alchemy_column(cls, alchemy_column: sqlalchemy.Column) -> DbColumnType:
        """
        Creates database column from underlying SQLAlchemy column
        :param sqlalchemy.Column alchemy_column: SQLAlchemy column definition
        :return: new DbColumn instance
        :rtype: DbColumn
        """
        return DbColumn(alchemy_column.name, alchemy_column)

    @classmethod
    def from_code(cls, column_name: str, column_code: Dict[str, Any]) -> DbColumnType:
        """
        Creates database column from passed SQLAlchemy column
        :param str column_name: column name
        :param Dict[str, Any] column_code: DbColumn's code definition
        :return: new DbColumn instance
        :rtype: DbColumn
        """
        fks = column_code.get('foreign_keys')
        foreign_keys = [ForeignKey(column=f) for f in fks] if fks is not None else []
        return DbColumn(column_name, Column(column_name, SqlParser.parse_alchemy_data_type(column_code['type']),
                                            primary_key=column_code.get('primary_key', False),
                                            nullable=column_code.get('nullable', True),
                                            # autoincrement=column_code.get('autoincrement'),
                                            unique=column_code.get('unique', False),
                                            index=column_code.get('index', False),
                                            default=column_code.get('default'),
                                            *foreign_keys
                                            ))

    @classmethod
    def validate_code(cls, column_name: str, column_code: Dict[str, Any]) -> Tuple[bool, Dict[str, Any]]:
        """
        Validates column code
        :param str column_name: column code
        :param Dict[str, Any] column_code: column code
        :return: True if column is valid, False otherwise and Dict of errors
        :rtype: Tuple[bool, Dict[str, Any]]
        """
        column_errors = {}
        # TODO: validate the column_name
        column_validator = Validator(
            {
                'type': {'type': 'string'},
                'primary_key': {'type': 'boolean'},
                'nullable': {'type': 'boolean'},
                'unique': {'type': 'boolean'},
                'index': {'type': 'boolean'},
                'default': {'type': ['string', 'boolean', 'date', 'datetime', 'float', 'integer', 'number']},
                'foreign_keys': {'type': 'list'}
            })
        validation_result = column_validator.validate(column_code)
        if not validation_result:
            column_errors['structure'] = column_validator.errors
        fk_errors = []
        for foreign_key_column in column_code.get('foreign_keys', []):
            fk_validation_result = (len(foreign_key_column.split('.')) == 2)
            if not fk_validation_result:
                fk_errors.append({
                    foreign_key_column: f"Invalid format ( not <table>.<column>)."})
            validation_result = validation_result and fk_validation_result
        if len(fk_errors) > 0:
            column_errors.update({'foreign_keys': fk_errors})
        return validation_result, column_errors



from jinja2 import Environment, PackageLoader
from sqlalchemy import PrimaryKeyConstraint, ForeignKeyConstraint, CheckConstraint, UniqueConstraint

import dbd
from dbd.utils.text_utils import strip_table_name

JINJA_GENERATOR_ENV = Environment(loader=PackageLoader(dbd.__name__, 'generator/generator_templates'))
JINJA_GENERATOR_ENV.globals.update(
    len=len, str=str, type=type, isinstance=isinstance,
    strip_table_name=strip_table_name,
    PrimaryKeyConstraint=PrimaryKeyConstraint,
    ForeignKeyConstraint=ForeignKeyConstraint,
    CheckConstraint=CheckConstraint,
    UniqueConstraint=UniqueConstraint
)



import logging
import os
import re
from os.path import exists
from typing import List, Dict, Any

import click
import networkx as nx
import sqlalchemy.engine
import yaml
from jinja2 import Environment, FileSystemLoader
from sqlalchemy import MetaData

from dbd.config.dbd_project import DbdProject
from dbd.log.dbd_exception import DbdException
from dbd.tasks.data_task import DataTask
from dbd.tasks.db_table_task import DbTableTask
from dbd.tasks.ddl_task import DdlTask
from dbd.tasks.elt_task import EltTask
from dbd.tasks.task import Task
from dbd.utils.sql_parser import SqlParser
from dbd.utils.text_utils import remove_prefix, relative_path_to_base_dir

log = logging.getLogger(__name__)


class ModelExecutionException(DbdException):
    pass


class InvalidModelException(DbdException):
    pass


# noinspection PyMethodMayBeStatic
class ModelExecutor:
    """
    Represents a top level data model blueprint that includes all elements (schema, tables, ELT operations, tests, etc.)
    The model is represented by multiple code files located in a hierarchy of directories on disk.
    """

    def __init__(self, project: DbdProject):
        """
        Constructor
        :param DbdProject project: DBD project
        """
        self.__project = project
        self.__model_directory = project.model_directory_from_project()
        self.__tasks = {}
        self.__ddl_tasks = {}
        self.__metadata_cache = {}

        self.__jinja_model_env = Environment(loader=FileSystemLoader(self.__model_directory))
        self.__jinja_model_env.globals.update()

    def reflect_metadata_cache(self, alchemy_engine: sqlalchemy.engine.Engine):
        """
        Reflects all cached metadata. This is necessary for detecting deletions
        :param sqlalchemy.engine.Engine alchemy_engine: SQLAlchemy engine database connection (both target and source)
        """
        # TODO: this approach can be slow - need to figure out, how to rebuild the cache incrementally
        log.debug(f"Reflecting metadata cache.")
        self.__build_metadata_cache(alchemy_engine)

    def validate(self):
        """
        Executes files stored in a model directory.
        The execution is performed in the database connected
        via the SQLAlchemy engine.
        """
        self.__populate_model_from_directory()
        ordered_tasks = self.__order_tasks_by_dependencies()
        task_errors = {}
        validation_result = True
        for task in reversed(ordered_tasks):
            log.debug(f"Validating task: '{task.task_id()}'.")
            task_validation_result, task_validation_errors = task.validate()
            if not task_validation_result:
                task_errors[task.task_id()] = task_validation_errors
                validation_result = False
            log.debug(f"Task validation finished: '{task.task_id()}'.")
        return validation_result, task_errors

    def execute(self, alchemy_engine: sqlalchemy.engine.Engine):
        """
        Executes files stored in a model directory. The execution is performed in the database connected via
        SQLAlchemy engine.
        :param sqlalchemy.engine.Engine alchemy_engine: SQLAlchemy engine database connection (both target and source)
        """
        try:
            self.__populate_model_from_directory()
            self.__build_metadata_cache(alchemy_engine)

            ordered_tasks = self.__order_tasks_by_dependencies()
            self.__drop_tables(ordered_tasks, alchemy_engine)
            self.reflect_metadata_cache(alchemy_engine)
            for task in reversed(ordered_tasks):
                schema = task.target_schema() if task.target_schema() is not None else Task.TOP_LEVEL_SCHEMA_NAME
                click.echo(f"Executing task: '{task.task_id()}'.")
                log.debug(f"Executing task: '{task.task_id()}'.")
                task.create(self.__metadata_cache[schema], alchemy_engine)
                log.debug(f"Task execution finished: '{task.task_id()}'.")
        except sqlalchemy.exc.OperationalError as o:
            log.error(f"Can't execute model because of: '{o}'.")
            raise ModelExecutionException(f"Can't execute model because of: '{o}'.")

    def __build_metadata_cache(self, alchemy_engine: sqlalchemy.engine.Engine):
        # noinspection GrazieInspection
        """
        Builds cache of SQLAlchemy MetaData objects. There is one MetaData object per schema.
        :param sqlalchemy.engine.Engine alchemy_engine: SQLAlchemy engine / database connection (both target and source
        schemas) for database objects creation, ELT tasks, etc.
        """
        # distinct schemas
        schemas = list(set([s.target_schema() for s in self.__tasks.values() if s.target_schema() is not None]))
        log.debug(f"Processing schemas: '{schemas}'.")
        for s in schemas:
            m = MetaData(bind=alchemy_engine, schema=s if s is not None and len(s) > 0 else None)
            m.reflect(views=True)
            self.__metadata_cache[s] = m
            log.debug(f"Metadata cache for schema {s} added.")
        # MetaData for the top-level objects and DDL
        m = MetaData(bind=alchemy_engine, schema=None)
        m.reflect(views=True)
        self.__metadata_cache[Task.TOP_LEVEL_SCHEMA_NAME] = m
        log.debug(f"Metadata cache for top level schema added.")

    def __apply_template(self, file: str, params: Dict[Any, Any]) -> str:
        """
        Applies Jinja2 template
        :param str file: Jinja2 template file
        :param Dict[Any, Any] params: parameters
        :return: processed content
        :rtype: str
        """
        template = self.__jinja_model_env.get_template(file)
        return template.render(params)

    def __process_data_file(self, model_root: str, dir_name: str, file_name: str, file_extension: str):
        """
        Process CSV file. Creates EltTask
        :param str model_root: model root directory
        :param str dir_name: schema directory
        :param str file_name: data file
        :param str file_extension: data file extension
        """
        file_name_absolute = relative_path_to_base_dir(self.__model_directory, model_root, file_name, file_extension)
        log.debug(f"Processing CSV file '{file_name_absolute}'.")
        template_params = dict(schema=dir_name, table=file_name, session=self)
        # do we have corresponding yaml config file?
        task_def = self.__load_yaml_metadata(model_root, dir_name, file_name, template_params)
        task = DataTask.from_code(task_def)
        task.set_task_data(os.path.join(self.__model_directory, file_name_absolute))
        self.__tasks[task.task_id()] = task
        log.debug(f"Added CSV file task: task_id='{task.task_id()}', task_data='{task.task_data()}'.")

    def __process_sql_file(self, model_root: str, dir_name: str, file_name: str, file_extension: str):
        """
        Process SQL (usually INSERTs from SELECTs) file. Creates EltTask
        :param str model_root: model root directory
        :param str dir_name: schema directory
        :param str file_name: SQL file
        :param str file_extension: SDL file extension
        """

        file_name_absolute = relative_path_to_base_dir(self.__model_directory, model_root, file_name, file_extension)
        log.debug(f"Processing SQL file '{file_name_absolute}'.")
        template_params = dict(schema=dir_name, table=file_name, session=self)
        processed_file = self.__apply_template(file_name_absolute, template_params)
        sql = SqlParser.compact_sql(processed_file)
        # do we have corresponding yaml config file?
        task_def = self.__load_yaml_metadata(model_root, dir_name, file_name, template_params)
        task = EltTask.from_code(task_def)
        task.set_task_data(sql)
        self.__tasks[task.task_id()] = task
        log.debug(f"Added SQL file task: task_id='{task.task_id()}', task_data='{task.task_data()}'.")

    def __process_ddl_file(self, model_root: str, dir_name: str, file_name: str, file_extension: str):
        """
        Process DDL (SQL DDL & DML statements)  file. Creates DdlTask
        :param str model_root: model root directory
        :param str dir_name: schema directory
        :param str file_name: DDL file
        :param str file_extension: DDL file extension
        """
        file_name_absolute = relative_path_to_base_dir(self.__model_directory, model_root, file_name, file_extension)
        log.debug(f"Processing DDL file '{file_name_absolute}'.")
        template_params = dict(schema=dir_name, table=file_name, session=self)
        processed_file = self.__apply_template(file_name_absolute, template_params)
        sql_without_comments = SqlParser.remove_sql_comments(processed_file)
        ddl = re.split(r';\s*$', sql_without_comments, flags=re.MULTILINE)

        task = DdlTask.from_code(file_name, dir_name)
        task.set_task_data(ddl)
        self.__ddl_tasks[task.task_id()] = task
        log.debug(f"Added DDL file task: task_id='{task.task_id()}', task_data='{task.task_data()}'.")

    def __load_yaml_metadata(self, model_root: str, dir_name: str, file_name: str, template_params: Dict[Any, Any]):
        """
        Loads corresponding YAML file with metadata, apply Jinja2 and extracts the table definition
        :param str model_root: model root directory
        :param str dir_name: YAML file directory
        :param str file_name: YAML file name
        :param Dict[Any, Any] template_params: Jinja2 context
        :return Dict[str, Any]: task definition
        """
        yaml_file_name_absolute = relative_path_to_base_dir(self.__model_directory, model_root, file_name, 'yaml')
        task_def = dict(runtime=dict(table=file_name, schema=dir_name))
        if exists(os.path.join(self.__model_directory, yaml_file_name_absolute)):
            log.debug(f"YAML config found '{os.path.join(self.__model_directory, yaml_file_name_absolute)}'.")
            processed_yaml = self.__apply_template(yaml_file_name_absolute, template_params)
            config = yaml.safe_load(processed_yaml)
            task_def.update(config)
            log.debug(f"Task definition enriched to task_def='{task_def}'.")
        return task_def

    def __populate_model_from_directory(self):
        """
        Crawls the model directory and collects all code files to a directory
        hashed by a target's name (e.g. database table name)
        """
        for model_root, dirs, files in os.walk(os.path.dirname(f"{self.__model_directory}/"), topdown=True):
            for filename in files:
                dir_name = f"{remove_prefix(model_root, self.__model_directory)}".split('/')[-1]
                schema_name = dir_name if len(dir_name) > 0 else None
                file_name, extension = os.path.splitext(filename)
                if extension.lower() == '.sql':
                    self.__process_sql_file(model_root, schema_name, file_name, extension)
                if extension.lower() == '.ddl':
                    self.__process_ddl_file(model_root, schema_name, file_name, extension)
                elif extension.lower() in ['.csv', '.json', '.xls', '.xlsx', '.xlsm', '.xlsb', '.odf', '.ods', '.odt',
                                           '.parquet']:
                    self.__process_data_file(model_root, schema_name, file_name, extension)
                elif extension.lower() == '.yaml':
                    pass

    def __drop_tables(self, tasks_ordered_by_dependencies: List[Task], alchemy_engine: sqlalchemy.engine.Engine):
        # noinspection GrazieInspection
        """
                Drops db tables as array of table names. The array should be ordered to satisfy the dependencies.
                :param List[str] tasks_ordered_by_dependencies: array of table names ordered in a way that
                    they can be dropped sequentially
                :param sqlalchemy.engine.Engine alchemy_engine: Engine SQLAlchemy engine database connection
                (both target and source schemas) for database objects creation, ELT tasks, etc.
                """
        for task in tasks_ordered_by_dependencies:
            schema = task.target_schema() if task.target_schema() is not None else Task.TOP_LEVEL_SCHEMA_NAME
            if isinstance(task, DbTableTask):
                # TODO: decide whether the default=drop is a good idea
                mode = task.process_def().get('mode', 'drop')
                if mode == 'drop':
                    log.debug(f"Dropping task with task_id='{task.task_id()}'.")
                    task.drop(self.__metadata_cache[schema], alchemy_engine)
                    log.debug(f"Dropped task with task_id='{task.task_id()}'.")
                elif mode == 'truncate':
                    log.debug(f"Truncating task with task_id='{task.task_id()}'.")
                    task.truncate(self.__metadata_cache[schema], alchemy_engine)
                    log.debug(f"Truncated task with task_id='{task.task_id()}'.")

    def __find_task_by_fully_qualified_target_name(self, fully_qualified_target_name):
        """
        Finds a task with target matching a fully qualified name
        :param fully_qualified_target_name: fully qualified target name
        :return: task with the target with matching name
        """
        found = [t for t in self.__tasks.values() if t.fully_qualified_target() == fully_qualified_target_name]
        if len(found) < 1:
            raise InvalidModelException(
                f"Invalid model: task with fully qualified name '{fully_qualified_target_name}' "
                f"not found!")
        elif len(found) > 1:
            raise InvalidModelException(
                f"Invalid model: multiple tasks with fully qualified name '{fully_qualified_target_name}' "
                f"found!")
        return found[0]

    def __order_tasks_by_dependencies(self) -> List[Task]:
        """
        Order tables by dependencies, so they can be created or dropped sequentially

        :return: list of tasks ordered in a way that satisfies the dependencies
        :rtype: List[Task]
        """
        log.debug(f"Ordering tasks '{self.__tasks.keys()}'.")
        dag_edges = []
        # order task by dependencies (items with most dependencies are at the beginning)
        for task in self.__tasks.values():
            task_table_dependencies = task.depends_on()
            for table_dependency in task_table_dependencies:
                try:
                    task_dependency = self.__find_task_by_fully_qualified_target_name(table_dependency)
                    dag_edges.append((task.task_id(), task_dependency.task_id()))
                except InvalidModelException:
                    # Some dependencies aren't resolved. We expect that these already exist
                    pass
        graph = nx.DiGraph()
        graph.add_edges_from(dag_edges)
        ordered_task_ids = list(nx.topological_sort(graph))
        ordered_tasks = [self.__tasks[tid] for tid in ordered_task_ids]
        independent_tasks = []
        for task in self.__tasks.values():
            if task not in ordered_tasks:
                independent_tasks.append(task)
        dag_order = independent_tasks + ordered_tasks
        # now we need to place the DDL tasks prolog at the end of each schema, epilog at the beginning
        # first remove the global epilog and prolog if they exist
        global_prolog = self.__ddl_tasks.pop(
            f"{Task.TOP_LEVEL_SCHEMA_NAME}{Task.TASK_ID_DELIMITER}{Task.TASK_TARGET_PROLOG}", None)
        global_epilog = self.__ddl_tasks.pop(
            f"{Task.TOP_LEVEL_SCHEMA_NAME}{Task.TASK_ID_DELIMITER}{Task.TASK_TARGET_EPILOG}", None)
        # Now global epilog and prolog are gone, we are with schema epilogs and prologs
        for epilog in [x for x in self.__ddl_tasks.values() if Task.TASK_TARGET_EPILOG == x.target()]:
            epilog_schema = epilog.target_schema()
            first_schema_index = self.find_first_task_for_schema(dag_order, epilog_schema)
            if first_schema_index >= 0:
                dag_order.insert(first_schema_index, epilog)
            else:
                dag_order.insert(0, epilog)

        for prolog in [x for x in self.__ddl_tasks.values() if Task.TASK_TARGET_PROLOG == x.target()]:
            prolog_schema = prolog.target_schema()
            last_schema_index = self.find_last_task_for_schema(dag_order, prolog_schema)
            if last_schema_index >= 0:
                dag_order.insert(last_schema_index + 1, prolog)
            else:
                dag_order.insert(len(dag_order), prolog)

        # global epilogs and prologs to the very beginning / end
        if global_prolog:
            dag_order.append(global_prolog)
        if global_epilog:
            dag_order.insert(0, global_epilog)
        log.debug(f"Task order '{[t.task_id() for t in dag_order]}'.")
        return dag_order

    def find_first_task_for_schema(self, tasks: List[Task], schema: str) -> int:
        """
        Finds the first task for schema
        :param List[Task] tasks: list of tasks
        :param str schema: schema name
        :return: the first index of task with the schema
        :rtype: int
        """
        for i in range(len(tasks)):
            if schema == tasks[i].target_schema():
                return i
        return -1

    def find_last_task_for_schema(self, tasks: List[Task], schema: str) -> int:
        """
        Finds the last task for schema
        :param List[Task] tasks: list of tasks
        :param str schema: schema name
        :return: the first index of task with the schema
        :rtype: int
        """
        for i in reversed(range(len(tasks))):
            if schema == tasks[i].target_schema():
                return i
        return -1



import logging
import re
import sys
import sqlalchemy.dialects.postgresql
from typing import List

import sqlalchemy
from sql_metadata import Parser

from dbd.log.dbd_exception import DbdException

log = logging.getLogger(__name__)


class SQlParserException(DbdException):
    pass


class SqlParser:
    """ Parses SQL and extracts different parts from the parsed SQL statement."""

    @classmethod
    def extract_tables(cls, sql: str) -> List[str]:
        """
        Extracts tables from the parsed SQL statement.
        Works with subquery, CTE, and many other SQL constructs
        :param str sql: parsed SQL statement
        :return: list of tables that the SQL statement depends on
        :rtype: List[str]
        """
        try:
            tables = Parser(sql).tables
        except ValueError:
            raise SQlParserException(f"Invalid SQL query '{sql}'")
        return tables

    @classmethod
    def extract_foreign_key_tables(cls, foreign_keys_def: List[str]) -> List[str]:
        """
        Extracts tables that the passed foreign keys depend on
        Relies on the <table>.<column> format
        :param List[str] foreign_keys_def: foreign key
        :return: str array of table names extracted from passed foreign keys
        :rtype: List[str]
        """
        for f in foreign_keys_def:
            if len(f.split('.')) <= 1:
                raise SQlParserException(f"Invalid foreign key format '{f}'. There is no table component.")
        return [f.split('.')[0] for f in foreign_keys_def]

    @classmethod
    def compact_sql(cls, sql: str) -> str:
        """
        Compacts the SQL text. Strip comments, etc.
        :param sql: input SQL
        :return: compacted SQL text
        :rtype: str
        """
        try:
            parsed_sql = Parser(sql).without_comments
        except ValueError:
            raise SQlParserException(f"Invalid SQL query '{sql}'")
        return parsed_sql

    @classmethod
    def comments(cls, sql: str) -> List[str]:
        """
        Returns the SQL comments
        :param sql: input SQL
        :return: comments as array of string
        :rtype: List[str]
        """
        try:
            parsed_sql = Parser(sql).comments
        except ValueError:
            raise SQlParserException(f"Invalid SQL query '{sql}'")
        return parsed_sql

    @classmethod
    def parse_alchemy_data_type(cls, data_type: str) -> sqlalchemy.types.TypeEngine:
        """
        Parses SQLAlchemy datatype from string
        :param str data_type: SQLAlchemy data type as string
        :return: SQLAlchemy data type
        :rtype:  sqlalchemy.types.TypeEngine subclass
        """
        parsed_data_type = Parser(f"CREATE TABLE a( c {data_type} )")
        # parsed_data_type = Parser(data_type)
        core_data_type = parsed_data_type.tokens[5].value
        length = int(parsed_data_type.tokens[7].value) if len(
            parsed_data_type.tokens) > 7 and parsed_data_type.tokens[7].is_integer else None
        scale = int(parsed_data_type.tokens[9].value) if len(
            parsed_data_type.tokens) > 9 and parsed_data_type.tokens[9].is_integer else None

        try:
            datatype_class = getattr(sys.modules['sqlalchemy.sql.sqltypes'], core_data_type)
        except AttributeError:
            try:
                datatype_class = getattr(sys.modules['sqlalchemy.dialects.postgresql'], core_data_type)
            except AttributeError:
                log.debug(f"Unsupported data type {core_data_type}.")
                raise SQlParserException(f"Unsupported data type {core_data_type}.")

        if core_data_type in ('CHAR', 'VARCHAR'):
            return datatype_class(length=length)
        elif core_data_type in ('DECIMAL', 'NUMERIC'):
            return datatype_class(precision=length, scale=scale)
        elif core_data_type in ('TIMESTAMP', 'DATE', 'INTEGER', 'FLOAT', 'DOUBLE', 'TEXT', 'SMALLINT',
                                'DOUBLE_PRECISION', 'REAL'):
            return datatype_class()
        else:
            log.debug(f"Unsupported data type {core_data_type}.")
            raise SQlParserException(f"Unsupported data type {core_data_type}.")

    @classmethod
    def remove_sql_comments(cls, sql_text: str) -> str:
        """
        Remove SQL comments from a SQL text
        :param str sql_text: SQL texts with comments
        :return: SQL texts without comments
        :rtype: str
        """
        pattern = r"(\".*?\"|\'.*?\')|(/\*.*?\*/|--[^\r\n]*$)"
        # first group captures quoted strings (double or single)
        # second group captures comments (//single-line or /* multi-line */)
        regex = re.compile(pattern, re.MULTILINE | re.DOTALL)

        def _replacer(match):
            # if the 2nd group (capturing comments) is not None,
            # it means we have captured a non-quoted (real) comment string.
            if match.group(2) is not None:
                return ""  # so we will return empty to remove the comment
            else:  # otherwise, we will return the 1st group
                return match.group(1)  # captured quoted-string

        no_comments = regex.sub(_replacer, sql_text)
        # replace multiple newlines with one
        return re.sub(r'\n+', '\n', no_comments).strip()



import os.path
from typing import Dict

from jinja2 import Environment, FileSystemLoader


def apply_template(file: str, params: Dict[str, str]) -> str:
    """
    Applies Jinja2 template
    :param str file: Jinja2 template file
    :param Dict[str, str] params: parameters
    :return: processed content
    :rtype: str
    """
    jinja_profile_env = Environment(loader=FileSystemLoader(os.path.dirname(file)))
    jinja_profile_env.globals.update()
    template = jinja_profile_env.get_template(os.path.basename(file))
    return template.render(params)



import os
from os.path import exists
from typing import Dict, TypeVar, Any

import sqlalchemy.engine
import yaml

from dbd.config.dbd_profile import DbdProfile
from dbd.log.dbd_exception import DbdException
from dbd.utils.jinja_utils import apply_template

ENV_VARS = {key: str(value) for key, value in os.environ.items()}

DbdProjectType = TypeVar('DbdProjectType', bound='DbdProject')


class DbdProjectConfigException(DbdException):
    pass


class DbdProject:
    """
    Project contains all project-level parameters and options
    """

    def __init__(self, profile: DbdProfile, project_file_full_path: str, config: Dict[str, Any]):
        """
        Constructor
        :param DbdProfile profile: DBD Profile
        :param str project_file_full_path: Path to the project file
        :param Dict[str, Any] config: Config
        """

        if profile is None or not isinstance(profile, DbdProfile):
            raise DbdProjectConfigException("Invalid profile.")
        self.__profile = profile
        self.__project_file = project_file_full_path
        self.__project_directory = os.path.dirname(project_file_full_path)
        self.__config = config

    @classmethod
    def load(cls, profile: DbdProfile, project_file_name='dbd.project') -> DbdProjectType:
        """
        Loads profile config file
        :param DbdProfile profile: DBD Profile
        :param str project_file_name: Profile file name
        :return: new DbdProject
        :rtype: DbdProject
        """
        project_file_name_path = os.path.join('.', project_file_name)
        if exists(project_file_name_path):
            processed_yaml = apply_template(project_file_name_path, ENV_VARS)
            return DbdProject(profile, os.path.normpath(project_file_name_path), yaml.safe_load(processed_yaml))
        raise DbdProjectConfigException(
            f"Can't find DBD project file. Searched the current dir and your home dir for '{project_file_name_path}'")

    def alchemy_engine_from_project(self) -> sqlalchemy.engine.Engine:
        """
        Returns SQLAlchemy engine initialized from project parameters
        :return: SQLAlchemy engine initialized from project parameters
        :rtype: sqlalchemy.engine.Engine
        """
        if self.__config is not None and 'database' in self.__config:
            return self.__profile.alchemy_engine_from_profile(self.__config.get('database'))
        else:
            raise DbdProjectConfigException(f"Project file '{self.__project_directory}{os.sep}{self.__project_file}' "
                                            f"doesn't contain 'database' key.")

    def profile(self) -> DbdProfile:
        """
        Returns associated profile
        :return: associated profile
        :rtype: DbdProfile
        """
        return self.__profile

    def model_directory_from_project(self) -> str:
        """
        Returns model directory initialized from project
        :return: model directory initialized from project
        :rtype: sqlalchemy.engine.Engine
        """
        model_dir = os.path.join(self.__project_directory, self.__config.get(
            'model')) if self.__config is not None and 'model' in self.__config else './model'
        if exists(model_dir):
            return os.path.normpath(model_dir.rstrip(os.sep))
        else:
            raise DbdProjectConfigException(f"Can't find project file '{self.__project_file}' or it doesn't "
                                            f"contain 'model' key and no 'model' dir exists in your current dir.")



import os
from typing import Dict, List, Any, TypeVar

import pandas as pd
import sqlalchemy.engine
from sqlalchemy import Column, TEXT

from dbd.db.db_table import DbTable
from dbd.log.dbd_exception import DbdException
from dbd.tasks.db_table_task import DbTableTask


class UnsupportedDataFile(DbdException):
    pass


DataTaskType = TypeVar('DataTaskType', bound='DataTask')


class DataTask(DbTableTask):
    """
    Data loading task. Loads data from a local data file (e.g. CSV) to database.
    """

    def __init__(self, task_def: Dict[str, Any]):
        """
        Data task constructor
        :param Dict[str, Any] task_def: Target table definition
        """
        super().__init__(task_def)
        self.__data_frame = None

    def data_file(self) -> str:
        """
        Task SQL statement
        :return: task SQL statement
        """
        return self.task_data()

    def set_data_file(self, sql: str):
        """
        Sets task SQL
        :param sql: SQL statement
        """
        self.set_task_data(sql)

    @classmethod
    def from_code(cls, task_def: Dict[str, Any]) -> DataTaskType:
        """
        Creates a new task from table definition (dict)
        :param Dict[str, Any] task_def: table definition (dict)
        :return: new EltTask instance
        :rtype: EltTask
        """

        return DataTask(task_def)

    def __data_file_columns(self) -> List[sqlalchemy.Column]:
        """
        Introspects data file columns
        :return: list of data file columns (SQLAlchemy Column[])
        :rtype: List[sqlalchemy.Column]
        """
        return [Column(c, TEXT) for c in self.__data_frame.columns]

    # noinspection DuplicatedCode
    def __override_data_file_column_definitions(self) -> Dict[str, Any]:
        """
        Merges the data file column definitions with the column definitions from the task_def.
        The column definitions override the introspected data file types

        :return: data file columns overridden with the task's explicit column definitions
        :rtype: Dict[str, Any]
        """
        table_def = self.table_def()
        column_overrides = table_def.get('columns', {})
        data_file_columns = self.__data_file_columns()
        ordered_columns = {}
        for c in data_file_columns:
            overridden_column = column_overrides.get(c.name)
            if overridden_column:
                if 'type' not in overridden_column:
                    overridden_column['type'] = c.type
                ordered_columns[c.name] = overridden_column
            else:
                ordered_columns[c.name] = {"type": c.type}

        table_def['columns'] = ordered_columns
        return table_def

    def create(self, target_alchemy_metadata: sqlalchemy.MetaData, alchemy_engine: sqlalchemy.engine.Engine):
        """
        Executes the task. Creates the target table and loads data
        :param sqlalchemy.MetaData target_alchemy_metadata: MetaData SQLAlchemy MetaData
        :param sqlalchemy.engine.Engine alchemy_engine:
        """
        self.__read_file_to_dataframe()

        table_def = self.__override_data_file_column_definitions()
        db_table = DbTable.from_code(self.target(), table_def, target_alchemy_metadata, self.target_schema())
        self.set_db_table(db_table)
        db_table.create()

        self.__data_frame.to_sql(self.target(), alchemy_engine, schema=self.target_schema(),
                                 if_exists='append', index=False)

    def __read_file_to_dataframe(self):
        """
        Read the file content to Pandas dataframe
        """
        absolute_file_name = self.data_file()
        file_name, file_extension = os.path.splitext(absolute_file_name)
        if file_extension.lower() == '.csv':
            self.__data_frame = pd.read_csv(absolute_file_name)
        elif file_extension.lower() == '.json':
            self.__data_frame = pd.read_json(absolute_file_name)
        elif file_extension.lower() in ['.xls', '.xlsx', '.xlsm', '.xlsb', '.odf', '.ods', '.odt']:
            self.__data_frame = pd.read_excel(absolute_file_name)
        elif file_extension.lower() == '.parquet':
            self.__data_frame = pd.read_parquet(absolute_file_name)
        else:
            raise UnsupportedDataFile(f"Data files with extension '{file_extension}' aren't supported.")



from typing import Dict, List, Any, TypeVar, Tuple

import sqlalchemy.engine
from cerberus import Validator

from dbd.db.db_table import DbTable
from dbd.tasks.task import Task, InvalidTaskDefinition
from dbd.utils.sql_parser import SqlParser
from dbd.utils.text_utils import fully_qualified_table_name

DbTableTaskType = TypeVar('DbTableTaskType', bound='DbTableTask')


class DbTableTask(Task):
    """
    Generic DB table task. Its result is a DB table that sits in a schema
    """

    def __init__(self, task_def: Dict[str, Any]):
        """
        DbTableTask constructor
        :param Dict[str, Any] task_def: table definition
        """

        self.__task_def = task_def
        self.__target_db_table = None

        runtime_def = self.runtime_def()

        if 'schema' not in runtime_def:
            raise InvalidTaskDefinition("Missing task runtime 'schema' key.")
        if 'table' not in runtime_def:
            raise InvalidTaskDefinition("Missing task runtime 'table' key.")

        table = runtime_def.get('table', None)
        schema = runtime_def.get('schema', None)
        task_id = Task.generate_task_id(table, schema)
        super().__init__(task_id, table, schema)

    def task_def(self) -> Dict[str, Any]:
        """
        Task definition dict
        :return: task definition (dict)
        :rtype: Dict[str, Any]
        """
        return self.__task_def

    def table_def(self) -> Dict[str, Any]:
        """
        Table definition dict
        :return: the table definition (dict)
        :rtype: Dict[str, Any]
        """
        return self.__task_def.get('table', {})

    def runtime_def(self) -> Dict[str, Any]:
        """
        Return runtime parameters
        :return: runtime parameters
        :rtype: Dict[str, Any]
        """
        return self.__task_def.get('runtime', {})

    def process_def(self) -> Dict[str, Any]:
        """
        Process definition dict
        :return: process definition (dict)
        :rtype: Dict[str, Any]
        """
        return self.__task_def.get('process', {})

    def db_table(self) -> DbTable:
        """
        Underlying DbTable
        :return: underlying DbTable
        :rtype: DbTable
        """
        return self.__target_db_table

    def set_db_table(self, db_table: DbTable):
        """
        Sets underlying DbTable
        :param DbTable db_table: DbTable
        """
        self.__target_db_table = db_table

    def drop(self, alchemy_metadata: sqlalchemy.MetaData, alchemy_engine: sqlalchemy.engine.Engine):
        """
        Drops the target table in the database
        :param sqlalchemy.MetaData alchemy_metadata: SqlAlchemy metadata
        :param sqlalchemy.engine.Engine alchemy_engine: Engine SQLAlchemy engine database connection
        """
        if self.__target_db_table is None:
            alchemy_table = alchemy_metadata.tables.get(self.fully_qualified_target())
            if alchemy_table is not None:
                self.__target_db_table = DbTable.from_alchemy_table(alchemy_table)
                self.__target_db_table.drop(alchemy_engine)
        else:
            self.__target_db_table.drop(alchemy_engine)

    def truncate(self, alchemy_metadata: sqlalchemy.MetaData, alchemy_engine: sqlalchemy.engine.Engine):
        """
        Truncates the target table
        :param sqlalchemy.MetaData alchemy_metadata: SqlAlchemy metadata
        :param sqlalchemy.engine.Engine alchemy_engine: Engine SQLAlchemy engine database connection
        """
        if self.__target_db_table is None:
            # TODO: fix the SQL statement to work for all DBs
            alchemy_table = alchemy_metadata.tables.get(self.fully_qualified_target())
            if alchemy_table is not None:
                self.__target_db_table = DbTable.from_alchemy_table(alchemy_table)
                self.__target_db_table.truncate(alchemy_engine)
        else:
            self.__target_db_table.truncate(alchemy_engine)

    def depends_on(self) -> List[str]:
        """
        Returns list of tables that the task depends on
        :return: list of tables that the task depends on
        :rtype: List[str]
        """
        target_db_schema = self.target_schema()
        columns = self.table_def().get('columns', {})
        foreign_key_tables = []
        for column in columns.values():
            column_foreign_keys = column.get('foreign_keys', [])
            foreign_key_tables += SqlParser.extract_foreign_key_tables(column_foreign_keys)
        return [f"{fully_qualified_table_name(target_db_schema, t)}" if len(t.split('.')) < 2 else t for t in
                foreign_key_tables]

    def validate(self) -> Tuple[bool, Dict[str, Any]]:
        """
        Validates task
        :return: True if task is valid, False otherwise and Dict of errors
        :rtype: Tuple[bool, Dict[str, Any]]
        """
        # TODO validate materialization, mode etc.
        task_errors = {}
        task_validator = Validator(
            {
                'table': {'type': 'dict'},
                'process': {'type': 'dict'},
                'runtime': {'type': 'dict'}
            })
        validation_result = task_validator.validate(self.__task_def)
        if not validation_result:
            task_errors['structure'] = task_validator.errors
        table_name = fully_qualified_table_name(self.target_schema(), self.target())
        table_validation_result, table_validation_errors = DbTable.validate_code(table_name, self.table_def())
        if not table_validation_result:
            task_errors['table'] = table_validation_errors
            validation_result = False

        process_validator = Validator(
            {
                'materialization': {'type': 'string'},
                'mode': {'type': 'string'}
            })
        process_validation_result = process_validator.validate(self.process_def())
        if not process_validation_result:
            task_errors['process'] = process_validator.errors
            validation_result = False

        runtime_validator = Validator(
            {
                'schema': {'type': 'string', 'nullable': True},
                'table': {'type': 'string'}
            })
        runtime_validation_result = runtime_validator.validate(self.runtime_def())
        if not runtime_validation_result:
            task_errors['runtime'] = runtime_validator.errors
            validation_result = False

        return validation_result, task_errors



import os
from os.path import exists
from pathlib import Path
from typing import Dict, Any, TypeVar

import sqlalchemy.engine
import yaml
from sqlalchemy import engine_from_config

from dbd.log.dbd_exception import DbdException
from dbd.utils.jinja_utils import apply_template

ENV_VARS = {key: str(value) for key, value in os.environ.items()}

CONFIG_PREFIX = 'db.'

DbdProfileType = TypeVar('DbdProfileType', bound='DbdProfile')


class DbdProfileConfigException(DbdException):
    pass


class DbdProfile:
    """
    Profile contains all generic parameters and options (e.g. database connections)
    """

    def __init__(self, profile_file_full_path: str, config: Dict[str, Any]):
        """
        Constructor
        :param str profile_file_full_path: Full path to the profile file
        :param Dict[str, str] config: Profile config
        """
        self.__profile_file = profile_file_full_path
        self.__config = config

    @classmethod
    def load(cls, profile_file_name='dbd.profile') -> DbdProfileType:
        """
        Loads profile config file
        :param str profile_file_name: Profile file name
        :return: new DbdProfile
        :rtype: DbdProfile
        """
        for profile_file_name in [os.path.join('.', profile_file_name),
                                  os.path.join(str(Path.home()), 'dbd.profile')]:
            if exists(profile_file_name):
                processed_yaml = apply_template(profile_file_name, ENV_VARS)
                return DbdProfile(os.path.normpath(profile_file_name), yaml.safe_load(processed_yaml))

        raise DbdProfileConfigException(
            f"Can't find DBD profile file. Searched the current dir and your home dir for '{profile_file_name}'")

    def db_connections(self) -> Dict[str, Dict[str, Any]]:
        """
        Return database connections from the profile
        :return: database connections from the profile
        :rtype: Dict[str, Dict[str, Any]]
        """
        if self.__config is not None:
            if 'databases' in self.__config:
                return self.__config.get('databases')
            else:
                raise DbdProfileConfigException(
                    f"Your dbd profile '{self.__profile_file}' doesn't contain 'database' key.")

    def alchemy_engine_from_profile(self, connection_name: str) -> sqlalchemy.engine.Engine:
        """
        Returns SQLAlchemy engine initialized from profile
        :param str connection_name: connection name
        :return: SQLAlchemy engine initialized from profile
        :rtype: sqlalchemy.engine.Engine
        """
        databases = self.db_connections()
        if connection_name in databases:
            try:
                return engine_from_config(databases.get(connection_name), prefix=CONFIG_PREFIX)
            except Exception:
                raise DbdProfileConfigException(f"Invalid connection '{connection_name}' config in the profile file "
                                                f"'{self.__profile_file}'.")
        else:
            raise DbdProfileConfigException(f"Connection '{connection_name}' isn't defined in the profile file "
                                            f"'{self.__profile_file}'.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dbd',
 'dbd.cli',
 'dbd.config',
 'dbd.db',
 'dbd.executors',
 'dbd.generator',
 'dbd.log',
 'dbd.tasks',
 'dbd.utils']

package_data = \
{'': ['*'],
 'dbd': ['resources/template/*', 'resources/template/model/*'],
 'dbd.generator': ['generator_templates/*']}

install_requires = \
['cerberus>=1.3.4,<2.0.0',
 'click>=8.0.3,<9.0.0',
 'jinja2>=3.0.3,<4.0.0',
 'networkx>=2.6.3,<3.0.0',
 'openpyxl>=3.0.9,<4.0.0',
 'pandas>=1.3.5,<2.0.0',
 'psycopg2>=2.9.3,<3.0.0',
 'pyarrow>=6.0.0,<7.0.0',
 'pyyaml>=6.0,<7.0',
 'sql-metadata>=2.3.0,<3.0.0',
 'sqlalchemy>=1.4.29,<2.0.0',
 'sqlparse>=0.4.2,<0.5.0']

entry_points = \
{'console_scripts': ['dbd = dbd.cli.dbdcli:cli']}

setup_kwargs = {
    'name': 'dbd',
    'version': '0.5.9',
    'description': 'Framework for declarative database creation and management.',
    'long_description': "# DBD - tool for declarative database definition\nDBD tool allows you to define your database schema and content declaratively. Database is represented by a \nhierarchy of directories and files stored in a model directory. \n\n![How DBD works](https://raw.githubusercontent.com/zsvoboda/dbd/master/img/dbd.infographic.png)\n\nDBD is great for declarative creation of any database. It is particularly designed for ELT (Extract/Load/Transform) \ndata pipelines used in data analytics and data warehousing. \n\n## TLDR: Whetting Your Appetite\n\n1. `dbd init test`\n2. `cd test`\n3. Check out files in the `model` directory.  \n4. `dbd validate .` \n5. `dbd run .`\n6. Connect to the newly created `states.db` sqlite database and review `area`, `population`, and `state` tables that have been created from the files in the `model` directory.\n\nNow you can delete the example files from the `model` directory, copy your Excel, JSON, or CSV files there instead. \nThen execute `dbd run .` again. Your files should be loaded in the `states.db` database.\n\nYou can create a YAML configuration files for your data (Excel, JSON, or CSV) files to specify individual column's\ndata types, indexes or constraints (e.g. primary key, foreign key, or check). See below for more details. \n\nYou can also add an SQL file that performs insert-from-select SQL statement to create database tables with transformed data.\n\n## Install DBD\nDBD requires Python 3.7.1 or higher. \n\n### PyPI\n\n```shell\npip3 install dbd\n```\n\nOR\n\n```shell\ngit clone https://github.com/zsvoboda/dbd.git\ncd dbd\npip3 install .\n```\n\n### Poetry\n\n```shell\ngit clone https://github.com/zsvoboda/dbd.git\ncd dbd\npoetry install\n``` \n\n## Generate a new DBD project\nYou can generate DBD project initial layout by executing `init` command:\n\n```shell\ndbd init <new-project-name>\n```\n\nThe `init` command generates a new DBD project directory with the following content: \n\n- `model` directory that contains the content files. dbd supports files with `.sql`, `.ddl`, `.csv`, `.json`, `.xlsx` and other extensions.  \n- `dbd.profile` configuration file that specifies database connections \n- `dbd.project` project configuration file\n\n## DBD profile configuration file\nDBD stores database connections in the `dbd.profile` configuration file. DBD searches for `dbd.profile` file in current or in \nyour home directory. You can always specify a custom profile file location using the `--profile` option of the `dbd` command. \n\nThe profile file is YAML file with the following structure:\n\n```yaml\ndatabases:\n  states:\n    db.url: <sql-alchemy-database-url>\n```\n\nRead more about [SQL Alchemy database URLs here](https://docs.sqlalchemy.org/en/14/core/engines.html). \n\nThe profile file can contain Jinja2 macros that substitute your environment variables. For example, you can reference \ndatabase password stored in a `SQLITE_PASSWORD` environment variable via `{{ SQLITE_PASSWORD }}` in your DBD profile.\n\n## DBD project configuration file\nDBD stores project configuration in a project configuration file that is usually stored in your DBD project directory. \nDBD searches for `dbd.project` file in your project's directory root. You can also use the `--project` option of the `dbd` \ncommand to specify a custom project configuration file. \n\nThe project configuration file also uses YAML format and references the DBD model directory with the `.sql`, `.csv` \nand other supported files. It also references the database configuration from the profile config file. For example:\n\n```yaml\nmodel: model\ndatabase: states\n```\n\nSimilarly like the profile file, you can use the environment variables substitution in the project config file too \n(e.g. `{{ SQLITE_DB_NAME }}`).\n\n## Model directory\nModel directory contains directories and DBD files. Each subdirectory of the model directory represents \na database schema. For example, this model directory structure\n\n```text\ndbd-project-directory\n+- schema1\n +-- us_states.csv\n+- schema2\n +-- us_counties.csv\n```\n\ncreates two database schemas: `schema1` and `schema2` and creates two database tables: `us_states` in `schema1` \nand `us_counties` in `schema2`. Both tables are populated with the data from the CSV files.  \n\nDBD supports following files located in the `model` directory:\n\n* __DATA files:__ `.csv`, `.json`, `.xls`, `.xlsx`, `.parquet` files are loaded to the database as tables\n* __SQL files:__ with SQL SELECT statements are executed using insert-from-select SQL construct. The INSERT command is generated (the SQL file only contains the SQL SELECT statement)\n* __DDL files:__ contain a sequence of SQL statements separated by semicolon. The DDL files can be named `prolog.ddl` and `epilog.ddl`. The `prolog.ddl` is executed before all other files in a specific schema. The `epilog.ddl` is executed last. The `prolog.ddl` and `epilog.ddl` in the top-level model directory are executed as the very first and tne very last files in the model. \n* __YAML files:__ specify additional configuration to the __DATA__ and __SQL__ files.\n\n## SQL files \nSQL file performs SQL transformation within your database. It contains a SQL SELECT statement that DBD wraps in \ninsert-from-select statement, executes it, and stores the result into a table or view that inherits its name from the \nSQL file name.\n\nHere is an example of `us_states.sql` file that creates a new `us_states` database table.\n\n```sqlite\nSELECT\n        state.abbrev AS state_code,\n        state.state AS state_name,\n        population.population AS state_population,\n        area.area_sq_mi  AS state_area_sq_mi\n    FROM state\n        JOIN population ON population.state = state.abbrev\n        JOIN area on area.state_name = state.state\n```\n\n## YAML files\nYAML file specify additional configuration for a corresponding __DATA__ or __SQL__ file with the same base file name.\nHere is a YAML configuration example for the `us_states.sql` file above:\n\n```yaml\ntable:\n  columns:\n    state_code:\n      nullable: false\n      primary_key: true\n      type: CHAR(2)\n    state_name:\n      nullable: false\n      index: true\n      type: VARCHAR(50)\n    state_population:\n      nullable: false\n      type: INTEGER\n    state_area_sq_mi:\n      nullable: false\n      type: INTEGER\nprocess:\n  materialization: table\n  mode: drop\n```\n\nNote that we re-type the `state_population` and the `state_area_sq_mi` columns to INTEGER, disallow\nNULL values in all columns, and specify that the `state_code` column is table's primary key.\n\nThe table is dropped and data re-loaded in full everytime the dbd executes this model. \n\n### Table section\nYAML file's columns are mapped to a columns of the table that DBD creates from a corresponding DATA or SQL file. \nFor example, a CSV header or SQL SELECT column `AS` clause. You can specify the following column's parameters:\n\n* __type:__ column's SQL type.\n* __primary_key:__ is the column part of table's primary key (true|false)?\n* __foreign_keys:__ all other database table columns that are referenced from a column in table.column format\n* __nullable:__ does column allow null values (true|false)?\n* __index:__ is column indexed (true|false)?\n* __unique:__ does column store unique values (true|false)?\n\n### Process section\nThe `process` section specifies the following processing options:\n\n* __materialization:__ specifies whether DBD creates a physical `table` or a `view` when processing a SQL file.\n* __mode:__ specifies how DBD works with a table. You can specify values `drop`, `truncate`, or `keep`. The  __mode__ option is ignored for views.\n\n## License\nDBD code is open-sourced under [BSD 3-clause license](LICENSE). \n\n## Resources and References\n- [DBD github repo](https://github.com/zsvoboda/dbd)\n- [DBD PyPi](https://pypi.org/project/dbd/)\n- [Submit issue](https://github.com/zsvoboda/dbd/issues)\n",
    'author': 'zsvoboda',
    'author_email': 'zsvoboda@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/zsvoboda/dbd/blob/master/README.md',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Any, List, TypeVar, Tuple, Dict

import sqlalchemy

from dbd.log.dbd_exception import DbdException
from dbd.utils.text_utils import fully_qualified_table_name

TaskType = TypeVar('TaskType', bound='Task')


class InvalidTaskDefinition(DbdException):
    pass


class Task:
    """
    Generic task. Ancestor of all tasks that can be planned.
    """

    TOP_LEVEL_SCHEMA_NAME = '*'
    TASK_ID_DELIMITER = '.'
    TASK_TARGET_PROLOG = 'prolog'
    TASK_TARGET_EPILOG = 'epilog'

    @classmethod
    def generate_task_id(cls, target: str, schema: str):
        """
        Generate task ID
        :param str schema: target schema
        :param str target: target (usually a table)
        :return: task ID
        :rtype: str
        """
        return f"{schema if schema is not None else Task.TOP_LEVEL_SCHEMA_NAME}{Task.TASK_ID_DELIMITER}{target}"

    def __init__(self, task_id: str, target: str, target_schema: str):
        """
        Constructor
        :param str task_id: Task ID
        :param str target: Task target (usually a db table)
        :param str target_schema: Task target db schema
        """
        self.__task_id = task_id
        self.__target = target
        self.__target_schema = target_schema
        self.__data = None

    def task_id(self) -> str:
        """
        Returns task ID
        :return: task ID
        :rtype: str
        """
        return self.__task_id

    def target_schema(self) -> str:
        """
        Returns task target schema
        :return: task target schema
        :rtype: str
        """
        return self.__target_schema

    def target(self) -> str:
        """
        Returns task target (usually a db table) name
        :return: task target (usually a db table) name
        :rtype: str
        """
        return self.__target

    def fully_qualified_target(self) -> str:
        """
        Returns fully qualified task target (usually a db table) name
        :return: fully qualified task target (usually a db table) name
        :rtype: str
        """
        return fully_qualified_table_name(self.__target_schema, self.__target)

    def set_task_data(self, data: Any):
        """
        Sets task specific data (usually a SQL statement)
        :param Any data: task specific data (usually a SQL statement)
        """
        self.__data = data

    def task_data(self) -> Any:
        """
        :return: task specific data (usually a SQL statement)
        :rtype: Any (usually str)
        """
        return self.__data

    def depends_on(self) -> List[str]:
        """
        Returns list of dependencies (list of fully qualified names)
        :return: list of dependencies
        :rtype: List[str]
        """
        pass

    def __eq__(self, other: TaskType) -> bool:
        """
        Comparator
        :param Task other: another Task instance to compare this instance to
        :return: True if both Tasks are equal (have the same task_id), False otherwise
        :rtype: bool
        """
        if isinstance(other, Task):
            if self.task_id() == other.task_id():
                return True
        return False

    def validate(self) -> Tuple[bool, Dict[str, Any]]:
        """
        Validates task
        :return: True if schema is valid, False otherwise and Dict of errors
        :rtype: Tuple[bool, Dict[str, Any]]
        """
        return True, {}

    def create(self, target_alchemy_metadata: sqlalchemy.MetaData, alchemy_engine: sqlalchemy.engine.Engine):
        """
        Executes / creates the task
        :param sqlalchemy.MetaData target_alchemy_metadata: MetaData SQLAlchemy MetaData
        :param sqlalchemy.engine.Engine alchemy_engine:
        """
        pass

    def drop(self, alchemy_metadata: sqlalchemy.MetaData, alchemy_engine: sqlalchemy.engine.Engine):
        """
        Drops the target table in the database
        :param sqlalchemy.MetaData alchemy_metadata: SqlAlchemy metadata
        :param sqlalchemy.engine.Engine alchemy_engine: Engine SQLAlchemy engine database connection
        """
        pass



from typing import TypeVar, List, Any, Dict, Tuple

import sqlalchemy
from cerberus import Validator
from sqlalchemy import MetaData

from dbd.db.db_table import DbTable
from dbd.generator.jinja_generator_env import JINJA_GENERATOR_ENV

DbSchemaType = TypeVar('DbSchemaType', bound='DbSchema')


class DbSchema:
    """
    Database schema definition
    """

    def __init__(self, name: str, tables: List[DbTable], alchemy_metadata: sqlalchemy.MetaData):
        """
        DbSchema constructor
        :param str name: schema name
        :param List[DbTable] tables: DbTable[] schema tables
        :param sqlalchemy.MetaData alchemy_metadata: MetaData underlying SQLAlchemy MetaData
        """
        self.__name = name
        self.__alchemy_metadata = alchemy_metadata
        self.__tables = tables

    def __eq__(self, other: DbSchemaType) -> bool:
        """
        Comparator
        :param DbSchema other: another DbSchema instance to compare this instance to
        :return: True if both DbSchemas are equal, False otherwise
        :rtype: bool
        """
        if isinstance(other, DbSchema):
            my_tables = self.tables()
            other_tables = other.tables()
            for my_table in my_tables:
                other_table = [t for t in other_tables if t.name() == my_table.name()]
                if len(other_table) != 1 or my_table != other_table[0]:
                    return False
            return True
        return False

    def tables(self) -> List[DbTable]:
        """
        :return: DbTable[] array of tables
        :type: List[DbTable]
        """
        return self.__tables

    def table(self, name: str) -> DbTable:
        """
        Returns table by name
        :param name: table name
        :return: table (None if a table with such name doesn't exist)
        :rtype: DbTable
        """
        table_matches = [t for t in self.__tables if name == t.name()]
        return table_matches[0] if len(table_matches) == 1 else None

    def alchemy_metadata(self) -> sqlalchemy.MetaData:
        """
        :return: MetaData SQLAlchemy metadata
        :rtype: sqlalchemy.MetaData
        """
        return self.__alchemy_metadata

    def name(self) -> str:
        """
        :return: schema name
        :rtype: str
        """
        return self.__name

    def create(self):
        """
        TODO: Exception handling
        Generates SQL and creates all database objects (tables)
        :return: SQLAlchemy status
        """
        self.__alchemy_metadata.create_all()

    def code(self) -> str:
        """
        TODO: shall I turn this to JSON and to dict here?
        :return: DbSchema's code
        :rtype: str
        """
        template = JINJA_GENERATOR_ENV.get_template('schema.j2')
        return template.render(dict(sn=self.__name, ts=[t.alchemy_table() for t in self.__tables]))

    @classmethod
    def from_alchemy_engine(cls, schema_name: str, alchemy_engine: sqlalchemy.engine.Engine) -> DbSchemaType:
        """
        Creates a new schema from SQLAlchemy engine
        :param str schema_name: schema name
        :param sqlalchemy.engine.Engine alchemy_engine: SQLAlchemy engine
        :return: new DbSchema instance
        :rtype: DbSchema
        """
        alchemy_metadata = MetaData(bind=alchemy_engine, schema=schema_name)
        alchemy_metadata.reflect(views=True)
        return DbSchema(schema_name, [DbTable.from_alchemy_table(t) for t in alchemy_metadata.tables.values()],
                        alchemy_metadata)

    @classmethod
    def from_code(cls, schema_code: Dict[str, Any], alchemy_engine: sqlalchemy.engine.Engine) -> DbSchemaType:
        """
        Creates a new schema from schema code
        :param Dict[str, Any] schema_code: schema code
        :param sqlalchemy.engine.Engine alchemy_engine: Engine SQLAlchemy engine
        :return: new DbSchema instance
        :rtype: DbSchema
        """
        schema_name = schema_code.get('name', None)
        alchemy_metadata = MetaData(bind=alchemy_engine, schema=schema_name)
        alchemy_metadata.reflect(views=True)
        tables = []
        for t in [t for t in schema_code['tables'].items()]:
            table_name, table_code = t
            tables.append(DbTable.from_code(table_name, table_code, alchemy_metadata, schema=schema_name))
        return DbSchema(schema_name, tables, alchemy_metadata)

    @classmethod
    def validate_code(cls, schema_code: Dict[str, Any]) -> Tuple[bool, Dict[str, Any]]:
        """
        Validates schema  code
        :param Dict[str, Any] schema_code: schema code
        :return: True if schema is valid, False otherwise and Dict of errors
        :rtype: Tuple[bool, Dict[str, Any]]
        """
        schema_errors = {}
        schema_validator = Validator(
            {
                'tables': {'type': 'dict', 'required': True}
            })
        validation_result = schema_validator.validate(schema_code)
        if not validation_result:
            schema_errors['structure'] = schema_validator.errors
        tables_errors = []
        for t in [t for t in schema_code['tables'].items()]:
            table_name, table_code = t
            table_validation_result, table_validation_errors = DbTable.validate_code(table_name, table_code)
            if not table_validation_result:
                tables_errors.append({table_name: table_validation_errors})
                validation_result = False
        if len(tables_errors) > 0:
            schema_errors.update({'tables': tables_errors})
        return validation_result, schema_errors



import os


def remove_prefix(text: str, prefix: str) -> str:
    """
    Removes the prefix from the beginning of the text if the text starts with the prefix

    :param str text: text that potentially starts with a prefix
    :param str prefix: prefix that the text potentially starts with
    :return: the text with stripped prefix
    :rtype: str
    """
    if text.startswith(prefix):
        return text[len(prefix):]
    return text


def relative_path_to_base_dir(base_directory: str, directory: str, file_name: str, extension: str) -> str:
    """
    Returns path represented by the directory, file_name, and extension relative to the base_directory
    :param str base_directory: base directory
    :param str directory: directory
    :param str file_name: file name
    :param str extension: file extension
    :return: path represented by the directory, file_name, and extension relative to the base_directory
    :rtype: str
    """
    full_file_name = f"{os.path.join(directory, file_name)}{'' if extension.startswith('.') else '.'}{extension}"
    return remove_prefix(full_file_name, base_directory).lstrip(os.sep)


def strip_table_name(column_name: str) -> str:
    """
    Strips the table name from the fully qualified column name
    :param str column_name: column name in format <table>.<column_name>
    :return: the column name
    :rtype: str
    """
    return ','.split(column_name)[-1]


def fully_qualified_table_name(schema: str, table: str) -> str:
    """
    Returns fully qualified table name
    :param str schema: schema name
    :param str table: table name
    :return: fully qualified table name
    :rtype: str
    """
    if schema is None or len(schema) < 1:
        return table
    else:
        return f"{schema}.{table}"



from typing import List, Dict, TypeVar, Any, Tuple

import sqlalchemy.engine
from cerberus import Validator
from sqlalchemy import CheckConstraint, Table, PrimaryKeyConstraint, ForeignKeyConstraint, UniqueConstraint, Index, text
from sqlalchemy.exc import ProgrammingError

from dbd.db.db_column import DbColumn
from dbd.generator.jinja_generator_env import JINJA_GENERATOR_ENV
from dbd.utils.text_utils import fully_qualified_table_name

DbTableType = TypeVar('DbTableType', bound='DbTable')


class DbTable:
    """
    Database table definition
    """

    def __init__(self, name: str, columns: List[DbColumn], alchemy_table: sqlalchemy.Table):
        """
        DbTable constructor
        :param str name: table name
        :param List[DbColumn] columns:  table columns
        :param sqlalchemy.Table alchemy_table: Table underlying SQLAlchemy table
        """
        self.__name = name
        self.__columns = columns
        self.__alchemy_table = alchemy_table

    @staticmethod
    def __constraint_to_str(constraint: sqlalchemy.Constraint) -> str:
        # noinspection GrazieInspection
        """
                Returns constraint's string description for purposes of comparing it with other constraint
                :param sqlalchemy.Constraint constraint: SQLAlchemy table constraint (PrimaryKeyConstraint,
                ForeignKeyConstraint, UniqueConstraint, CheckConstraint)
                :return: the string description of the passed constraint
                :rtype: str
                """
        columns = f"{constraint.columns}" if hasattr(constraint, 'columns') else "[]"
        refcolumns = f"{constraint.refcolumns}" if hasattr(constraint, 'refcolumns') else "[]"
        sqltext = f"{constraint.sqltext}" if hasattr(constraint, 'sqltext') else "''"
        return f"{type(constraint)}(columns={columns}, refcolumns={refcolumns}, sqltext={sqltext})"

    def __eq__(self, other: DbTableType) -> bool:
        """
        Comparator
        :param DbTable other: another DbTable instance to compare this instance to
        :return: True if both DbTables are equal, False otherwise
        :rtype: bool
        """
        if isinstance(other, DbTable):
            if self.name() != other.name():
                return False
            my_columns = self.columns()
            other_columns = other.columns()
            for my_column in my_columns:
                other_column = [c for c in other_columns if c.name() == my_column.name()]
                if len(other_column) != 1 or my_column != other_column[0]:
                    return False
            return set([self.__constraint_to_str(c) for c in self.alchemy_table().constraints]) == set(
                [self.__constraint_to_str(c) for c in other.alchemy_table().constraints])
        return False

    def alchemy_table(self) -> sqlalchemy.Table:
        """
        :return: underlying SQLAlchemy table
        :rtype: sqlalchemy.Table
        """
        return self.__alchemy_table

    def columns(self) -> List[DbColumn]:
        """
        :return: table's columns
        :rtype: List[DbColumn]
        """
        return self.__columns

    def column(self, name: str) -> DbColumn:
        """
        Returns column by name
        :param name: column name
        :return: table (None if a column with such name doesn't exist)
        :rtype: DbTable
        """
        column_matches = [t for t in self.__columns if name == t.name()]
        return column_matches[0] if len(column_matches) == 1 else None

    def name(self) -> str:
        """
        :return: table name
        :rtype: str
        """
        return self.__name

    def code(self) -> str:
        """
        TODO: shall I turn this to JSON and to dict here?
        :return: DbTable's code
        :rtype: str
        """
        template = JINJA_GENERATOR_ENV.get_template('table.j2')
        return template.render(dict(t=self.__alchemy_table))

    def create(self):
        """
        TODO: exception handling
        Generates SQL and creates the table in the target database
        """
        self.__alchemy_table.create(checkfirst=True)

    def drop(self, alchemy_engine):
        """
        TODO: exception handling
        Drops the table in the target database. Also drops it if its a view.
        :param sqlalchemy.engine.Engine alchemy_engine: Engine SQLAlchemy engine database connection
        """
        try:
            self.__alchemy_table.drop(checkfirst=True)
        except ProgrammingError:
            # it may be view
            with alchemy_engine.connect() as conn:
                view_name = fully_qualified_table_name(self.__alchemy_table.schema, self.__alchemy_table.name)
                # TODO: make sure this works for all DBs
                conn.execute(text(f"DROP VIEW IF EXISTS {view_name}"))

    def truncate(self, alchemy_engine):
        """
        TODO: exception handling
        Truncates table data
        :param sqlalchemy.engine.Engine alchemy_engine: Engine SQLAlchemy engine database connection
        """
        with alchemy_engine.connect() as conn:
            table_name = fully_qualified_table_name(self.__alchemy_table.schema, self.__alchemy_table.name)
            # TODO: make sure this works for all DBs
            conn.execute(text(f"TRUNCATE TABLE {table_name}"))

    @classmethod
    def from_alchemy_table(cls, alchemy_table: sqlalchemy.Table) -> DbTableType:
        """
        Creates database table from passed SQLAlchemy table
        :param sqlalchemy.Table alchemy_table: SQLAlchemy table definition

        :return: new DbTable instance
        :rtype: DbTable
        """
        return DbTable(alchemy_table.name, [DbColumn.from_alchemy_column(c) for c in alchemy_table.columns.values()],
                       alchemy_table)

    @classmethod
    def from_code(cls, name: str, table_code: Dict[str, Any], alchemy_metadata: sqlalchemy.MetaData,
                  schema: str = None) -> DbTableType:
        """
        Creates database table from passed code
        :param str name: table name
        :param Dict[str, Any] table_code: DbTable's code definition
        :param sqlalchemy.MetaData alchemy_metadata: SQLAlchemy MetaData object
        :param str schema: table's schema
        :return: new DbTable instance
        """
        columns = cls.__extract_columns_from_table_code(table_code)
        constraints = cls.__extract_constraints_from_table_code(table_code)
        indexes = cls.__extract_indexes_from_table_code(name, table_code)

        arguments = [c.alchemy_column() for c in columns] + constraints + indexes
        return DbTable(name, columns, Table(name, alchemy_metadata, schema=schema, extend_existing=True, *arguments))

    @classmethod
    def __extract_indexes_from_table_code(cls, name: str, table_code: Dict[str, Any]) -> List[Index]:
        """
        Extracts index definitions from DbTable's code
        :param str name: table name for constructing default index name
        :param Dict[str, Any] table_code: DbTable code
        :return: Index array of SQLAlchemy indexes
        :rtype: List[Index]
        """
        indexes = []
        ti = table_code.get('indexes', {})
        for index in ti:
            cols = index.get('columns')
            unique = index.get('unique', False)
            index_name = index.get('name', f"idx_{name}_{len(indexes) + 1}")
            index_arguments = [index_name] + cols
            indexes.append(Index(*index_arguments, unique=unique))
        return indexes

    @classmethod
    def __extract_constraints_from_table_code(cls, table_code: Dict[str, Any]) -> List[sqlalchemy.Constraint]:
        """
        Extracts table constraint definitions from DbTable's code
        :param Dict[str, Any] table_code: DbTable code
        :return: array of SQLAlchemy constraints
        :rtype: List[sqlalchemy.Constraint]
        """
        constraints = []
        tc = table_code.get('constraints', {})
        for constraint in tc:
            constraint_type = constraint.get('type')
            cols = constraint.get('columns')
            refs = constraint.get('references')
            # TODO handle unknown constraint
            if constraint_type == 'checkConstraint':
                constraints.append(CheckConstraint(sqltext=constraint.get('sqltext')))
            elif constraint_type == 'primaryKeyConstraint':
                constraints.append(PrimaryKeyConstraint(*cols))
            elif constraint_type == 'foreignKeyConstraint':
                constraints.append(ForeignKeyConstraint(columns=cols, refcolumns=refs))
            elif constraint_type == 'uniqueConstraint':
                constraints.append(UniqueConstraint(*cols))
        return constraints

    @classmethod
    def __extract_columns_from_table_code(cls, table_code: Dict[str, Any]) -> List[DbColumn]:
        """
        Extracts table column definitions from DbTable's code
        :param Dict[str, Any] table_code: DbTable code
        :return: DbColumn array of table columns
        :rtype: List[DbColumn]
        """
        columns = []
        for c in table_code['columns'].items():
            column_name, column_code = c
            columns.append(DbColumn.from_code(column_name, column_code))
        return columns

    @classmethod
    def validate_code(cls, table_name: str, table_code: Dict[str, Any]) -> Tuple[bool, Dict[str, Any]]:
        """
        Validates table code
        :param str table_name: table name
        :param Dict[str, Any] table_code: table code
        :return: True if table is valid, False otherwise and Dict of errors
        :rtype: Tuple[bool, Dict[str, Any]]
        """
        table_errors = {}
        # TODO: validate the table_name
        table_validator = Validator(
            {
                'columns': {'type': 'dict'},
                'constraints': {'type': 'list'},
                'indexes': {'type': 'list'}
            })
        validation_result = table_validator.validate(table_code)
        if not validation_result:
            table_errors['structure'] = table_validator.errors
        column_errors = {}
        for c in [c for c in table_code.get('columns', {}).items()]:
            column_name, column_code = c
            columns_validation_result, column_validation_errors = DbColumn.validate_code(column_name, column_code)
            if not columns_validation_result:
                column_errors[column_name] = column_validation_errors
                validation_result = False
        if len(column_errors) > 0:
            table_errors.update({'columns': column_errors})
        constraint_errors = []
        for constraint_code in table_code.get('constraints', []):
            constraints_validation_result, constraints_validation_errors = cls.__validate_constraint(constraint_code)
            if not constraints_validation_result:
                constraint_errors.append(constraints_validation_errors)
                validation_result = False
        if len(constraint_errors) > 0:
            table_errors.update({'constraints': constraint_errors})
        index_errors = []
        for index_code in table_code.get('indexes', []):
            indexes_validation_result, indexes_validation_errors = cls.__validate_index(index_code)
            if not indexes_validation_result:
                index_errors.append(indexes_validation_errors)
                validation_result = False
        if len(index_errors) > 0:
            table_errors.update({'indexes': index_errors})
        return validation_result, table_errors

    @classmethod
    def __validate_index(cls, index_code: Dict[str, Any]) -> Tuple[bool, Dict[str, Any]]:
        """
        Validates index
        :param Dict[str, Any] index_code: index code
        :return: True if indexes are valid, False otherwise and Dict of errors
        :rtype: Tuple[bool, Dict[str, Any]]
        """
        index_errors = {}
        index_validator = Validator({
            'columns': {
                'type': 'list',
                'required': True
            },
            'unique': {'type': 'bool'}
        })
        index_validation_result = index_validator.validate(index_code)
        if not index_validation_result:
            return False, {'structure': index_validator.errors}
        return True, {}

    @classmethod
    def __validate_constraint(cls, constraint_code: Dict[str, Any]) -> Tuple[bool, Dict[str, Any]]:
        """
        Validates constraint
        :param Dict[str, Any] constraint_code: constraint code
        :return: True if constraints are valid, False otherwise and Dict of errors
        :rtype: Tuple[bool, Dict[str, Any]]
        """
        constraints_errors = {}
        constraint_validator = Validator({'type': {'allowed': ['checkConstraint', 'primaryKeyConstraint',
                                                               'foreignKeyConstraint', 'uniqueConstraint']}})
        constraint_validator.allow_unknown = True
        validation_result = constraint_validator.validate(constraint_code)
        if not validation_result:
            constraints_errors['structure'] = constraint_validator.errors
        constraint_type = constraint_code.get('type')
        if constraint_type == 'checkConstraint':
            check_constraint_validator = Validator({
                'type': {
                    'type': 'string',
                    'allowed': ['checkConstraint'],
                    'required': True
                },
                'sqltext': {
                    'type': 'string',
                    'required': True
                }
            })
            check_validation_result = check_constraint_validator.validate(constraint_code)
            if not check_validation_result:
                constraints_errors['checkConstraint'] = check_constraint_validator.errors
                validation_result = False
        elif constraint_type == 'primaryKeyConstraint':
            pk_constraint_validator = Validator({
                'type': {
                    'type': 'string',
                    'allowed': ['primaryKeyConstraint'],
                    'required': True
                },
                'columns': {
                    'type': 'list',
                    'required': True
                }
            })
            pk_validation_result = pk_constraint_validator.validate(constraint_code)
            if not pk_validation_result:
                constraints_errors['primaryKeyConstraint'] = pk_constraint_validator.errors
                validation_result = False
        elif constraint_type == 'foreignKeyConstraint':
            fk_constraint_validator = Validator({
                'type': {
                    'type': 'string',
                    'allowed': ['foreignKeyConstraint'],
                    'required': True
                },
                'columns': {
                    'type': 'list',
                    'required': True
                },
                'references': {
                    'type': 'list',
                    'required': True
                }
            })
            fk_validation_result = fk_constraint_validator.validate(constraint_code)
            if not fk_validation_result:
                constraints_errors['foreignKeyConstraint'] = fk_constraint_validator.errors
                validation_result = False
        elif constraint_type == 'uniqueConstraint':
            unique_constraint_validator = Validator({
                'type': {
                    'type': 'string',
                    'allowed': ['uniqueConstraint'],
                    'required': True
                },
                'columns': {
                    'type': 'list',
                    'required': True
                }
            })
            unique_validation_result = unique_constraint_validator.validate(constraint_code)
            if not unique_validation_result:
                constraints_errors['uniqueKeyConstraint'] = unique_constraint_validator.errors
                validation_result = False
        else:
            validation_result = False
        return validation_result, constraints_errors
